import string
import os
import re
import time
import httplib
import urllib
import urlparse
import base64
import Cookie
import MimeWriter
import StringIO

import ConfigParserExt

#
#   Result classes
#

class Result:
    """
        Hold on to the results of running a test or suite of tests.
        Implements the GoF Composite pattern.
    """
    def __init__( self
                , test
                , application=None
                , defaults={}
                , time_requests=1
                , check_responses=1
                , check_elapsed_times=1
                ):

        self._test = test
        self._application = application
        self._defaults = {}
        self._defaults.update( defaults )
        self._state = {}
        self._time_requests = time_requests
        self._check_responses = check_responses
        self._check_elapsed_times = check_elapsed_times and time_requests
        self._invocations = []
        self._errors = []
        self._fatal_errors = []
        self._children = []
        self._cookies = []
        self._parent = None

    #
    #   Queries
    #
    def getTest( self ):
        """
            Return our test object.
        """
        return self._test

    def getApplication( self ):
        """
            Return the root object of Zope.
        """
        return self._application

    def getDefaults( self ):
        """
            Return our test object.
        """
        result = {}
        result.update( self._defaults )
        return result

    def getParent( self ):
        """
            Return our parent request, or None.
        """
        return self._parent

    def setParent( self, parent ):
        """
            Setup our parent request.
        """
        self._parent = parent

    def getCookies( self ):
        """
            Return cookies we have accumulated during the request.
        """
        if self._cookies:
            return self._cookies
        #import pdb; pdb.set_trace()
        if self._parent:
            return self._parent.getCookies()
        return []

    def setCookies( self, cookies ):
        """
            Store our cookies.
        """
        self._cookies = cookies
        if self._parent:
            #import pdb; pdb.set_trace()
            self._parent.setCookies( cookies )

    def setStateValue( self, key, value ):
        """
            Remember a value (e.g., across invocations)
        """
        self._state[ key ] = value

    def getStateValue( self, key ):
        """
            Recall a value stored using 'setStateValue'.
        """
        return self._state[ key ]
    
    def timeRequests( self ):
        """
            Should we capture elapsed times?
        """
        return self._time_requests

    def checkResponses( self ):
        """
            Should we validate response values?
        """
        return self._check_responses
    
    def checkElapsedTimes( self ):
        """
            Should we validate elapsed times?
        """
        return self._check_elapsed_times
    
    def listInvocations( self, roll_up=1 ):
        """
            Return the list of invocations we manage directly, or
            transitively if 'roll_up' is true.
        """
        result = []
        result.extend( self._invocations )

        if roll_up:
            for child in self.listChildren():
                result.extend( child.listInvocations( roll_up ) )
        
        return result

    def listErrors( self, fatal_only=0, roll_up=1 ):
        """
            Return a list of fatal errors.
        """
        result = []
        if fatal_only:
            result.extend( self._fatal_errors )
        else:
            result.extend( self._fatal_errors )
            result.extend( self._errors )

        if roll_up:
            for child in self.listChildren():
                result.extend( child.listErrors( fatal_only, roll_up ) )

        return result

    def listChildren( self ):
        """
            Return our sub-results.
        """
        return tuple( self._children )

    def listDumpedResults( self ):
        result = []
        for invocation in self._invocations:
            result.append( invocation.dump() )
        return result

    #
    #   Mutators
    #
    def addInvocation( self, invocation, request ):
        """
            Remember a request invocation other than setup / postcondition.
        """
        self._invocations.append( invocation )
        self._validateInvocation( invocation, request )

    def newChild( self, test ):
        """
            Create, remember, and return a sub-result.
        """
        klass = self.__class__
        child = klass( test 
                     , self.getApplication()
                     , self.getDefaults()
                     , self.timeRequests()
                     , self.checkResponses()
                     , self.checkElapsedTimes()
                     )
        child.setParent( self )
        self._children.append( child )
        return child

    def logError( self, msg ):
        """
            Record a fatal error.
        """
        self._errors.append( msg )

    def logFatalError( self, msg ):
        """
            Record a fatal error.
        """
        self._fatal_errors.append( msg )

    def __call__( self ):
        return len( self._fatal_errors ) + len( self._errors ) == 0

    #
    #   Utilities
    #
    def _validateInvocation( self
                           , invocation
                           , request
                           , force_check_response=0
                           , force_check_time=0
                           ):

        rname = request.getName()

        if self.checkResponses() or force_check_response:
            expected = request.getExpectedResult()
            got = invocation.getResult()
            if got != expected:
                msg = '[%-10s] response: %s != %s' % ( rname , got, expected )
                self._fatal_errors.append( msg )

        elif self.checkElapsedTimes() or force_check_time:
            expected = request.getExpectedTime()
            got = invocation.getElapsedTime()
            if got != expected:
                msg = '[%-10s] elapsed: %s != %s' % ( rname, got, expected )
                self._errors.append( msg )
        
class InvocationBase:

    _start_time    = None
    _stop_time     = None

    def __init__( self, request ):
        self._request = request

    def getRequest( self ):
        return self._request

    def beginRequest( self ):
        self._start_time    = time.time()

    def endRequest( self ):
        self._stop_time     = time.time()

    def dump( self ):
        raise NotImplementedError

    def getResult( self ):
        raise NotImplementedError

    def getInterval( self ):
        return self._stop_time and self._stop_time - self._start_time or 0.0

class HTTPRequestInvocation( InvocationBase ):

    _reply_code    = 200
    _reply_message = None
    _reply_headers = None
    _reply_payload = None

    def update( self, reply_code, reply_message, reply_headers, reply_payload ):
        self._reply_code    = reply_code
        self._reply_headers = reply_headers
        self._reply_message = reply_message
        self._reply_payload = reply_payload

    def dump( self ):
        return ( 'HTTP Request: result = %s\n%s\n\n%s'
               % ( self.getReplyStatus()   
                 , self._reply_headers
                 , self._reply_payload
                 )
               )

    def getReplyCode( self ):
        return self._reply_code

    getResult = getReplyCode

    def getReplyStatus( self ):
        return '%d %s' % ( self._reply_code, self._reply_message )

    def getCookies( self ):
        """
            Return a sequence (maybe empty) of Cookie objects describing
            the cookies set by the response.
        """
        result = []

        for h in filter( lambda x: x.lower().startswith( 'set-cookie:' )
                       , self._reply_headers.headers ):

            cookie = Cookie.SmartCookie()
            cookie.load( h )
            result.append( cookie )

        return result

class ZEORequestInvocation( InvocationBase ):

    _result = None

    def update( self, result ):
        self._result = result

    def dump( self ):
        return 'ZEO Request: result = %s' % self._result

    def getResult( self ):
        return self._result


#
#   Request classes.
#

class RequestError( Exception ):
    pass


class RequestBase: # XXX scarecrow me?
    """
        Common base class for requests.
    """
    def __init__( self, name ):

        self._name = name
        self._expected_result = 1
        self._expected_time = 1.0       # second

    #
    #   Accessors
    #
    def getType( self ):
        """
            Return a type string identifying the kind of request
            (ick!).
        """
        raise RequestError, "Not implemented"

    def getName( self ):
        """
            Return the name of this request within the scenario.
        """
        return self._name

    def getExpectedResult( self ):
        """
            Return the expected result (return value, HTTP request code,
            etc.)
        """
        return self._expected_result

    def getExpectedTime( self ):
        """
            Return the maximum expected time to complete the request.
        """
        return self._expected_time

    def __call__( self, result ):
        """
            Invoke the request, storing results in 'result'.
        """
        raise NotImplementedError

class SleepRequest( RequestBase ):
    """
        "Spacer" request.
    """
    def __init__( self, name ):

        self._name = name
        self._expected_result = 1
        self._sleep_time = 1.0       # second

    #
    #   Accessors
    #
    def getType( self ):
        """
            Return a type string identifying the kind of request
            (ick!).
        """
        return 'SleepRequest'

    def getName( self ):
        """
            Return the name of this request within the scenario.
        """
        return self._name

    def getExpectedResult( self ):
        """
            Return the expected result (return value, HTTP request code,
            etc.)
        """
        return self._expected_result

    def getExpectedTime( self ):
        """
            Return the maximum expected time to complete the request.
        """
        return self._sleep_time

    def getSleepTime( self ):
        """
            Return the time to sleep, in seconds.
        """
        return self._sleep_time

    def setSleepTime( self, sleep_time ):
        """
            Set the sleep time, in seconds.
        """
        self._sleep_time = sleep_time

    def __call__( self, result ):
        """
            Invoke the request, storing results in 'result'.
        """
        import time
        time.sleep( self._sleep_time )

def _buildSleepRequest( cp, section ):
    """
        Construct a SleepRequest, using the values in 'cp' and 'section'.
    """
    r = SleepRequest( section )

    options = cp.options( section )

    if 'sleep' in options:
        r.setSleepTime( cp.getfloat( section, 'sleep' ) )

    return r


HKEY    = re.compile( r'Header_[0-9]+', re.IGNORECASE )
HEADER  = re.compile( r'([A-Za-z_][\w_]*)[ ]*=[ ]*(.*)' )

def _buildHeader( header_spec ):
    """
        Construct a '( key, value )' tuple representing an additional
        HTTP header.
    """
    match = HEADER.match( header_spec )
    if match:
        return match.groups()
    raise RequestError, 'Invalid header_spec: %s' % header_spec

CKEY    = re.compile( r'Cookie_[0-9]+', re.IGNORECASE )
COOKIE  = re.compile( r'([A-Za-z_][\w_]*)[ ]*=[ ]*(.*)' )

def _buildCookie( cookie_spec ):
    """
        Construct a '( key, value )' tuple representing a cookie to be
        sent with the request.
    """
    match = COOKIE.match( cookie_spec )
    if match:
        return match.groups()
    raise RequestError, 'Invalid cookie_spec: %s' % cookie_spec

FKEY    = re.compile( r'Field_[0-9]+', re.IGNORECASE )
FDESC   = re.compile( r'([A-Za-z_][\w_]*):(\w+)[ ]*=[ ]*(.*)' )
FDESC2  = re.compile( r'([A-Za-z_][\w_]*)[ ]*=[ ]*(.*)' )

def _buildField( field_desc ):
    """
        Construct a '( name, type, value )' tuple for a form field.

        'file' type fields should have the filename as 'value'.
    """
    match = FDESC.match( field_desc )
    if match:
        return match.groups()

    match = FDESC2.match( field_desc )
    if match:
        return match.group(1), None, match.group(2)

    raise RequestError, 'Invalid field_desc: %s' % field_desc

URLENCODED = 'application/x-www-form-urlencoded'
MULTIPART  = 'multipart/form-data'

def _mime_encode( fields ):
    """
        Build a 'multipart/form-data' representation of the
        (name,type,value) items in fields.
    """
    raise NotImplementedError

class HTTPRequest( RequestBase ):
    """
        Represent a single request/reply interaction with the system, 
        including both the parameters for invoking the system and the
        expected results.
    """

    def __init__( self, name, URL ):

        RequestBase.__init__( self, name )

        self._initURLParts( URL )

        self._authentication_token = None  # store as 'userid:password'
        self._headers = []
        self._cookies = []
        self._fields = []
        self._has_file_field  = 0
        self._expected_result = 200
        self._expected_redirect = None
        self._expected_cookies = ()
    
    def _initURLParts( self, URL ):

        parts = urlparse.urlparse( URL )
        self._scheme = parts[0]
        self._netloc = parts[1]
        hp = string.split( self._netloc, ':' )

        if len( hp ) > 2:
            raise ValueError, 'netloc "%s" malformed' % self._netloc

        if len( hp ) == 2:
            self._host, self._port = hp[0], int( hp[1] )
        else:
            self._host = self._netloc
            self._port = 80 # XXX s.b. default for scheme?

        self._path, self._params, self._query, self._fragment = parts[2:]
    #
    #   Accessors
    #
    def getType( self ):
        """
            Return the type, for sniffing purposes.
        """
        return 'HTTPRequest'

    def getMethod( self ):
        """
            Return the name of the HTTP method to use for this request.
        """
        return self._fields and 'POST' or 'GET'

    def getURI( self ):
        """
            Return the site-relative portion of the URL.
        """
        url = self._path
        if self._params:
            url = url + ';' + self._params
        if self._query:
            url = url + '?' + self._query
        if self._fragment:
            url = url + '#' + self._fragment
        return url


    def getURLScheme( self ):
        """
            Return the scheme (e.g., 'http', 'ftp') for the request.
        """
        return self._scheme

    def getHost( self ):
        """
            Return the hostname for the request.
        """
        return self._host

    def getPort( self ):
        """
            Return the port number for the request.
        """
        return self._port

    def getPath( self ):
        """
            Return the path for the request.
        """
        return self._path

    def getParameters( self ):
        """
            Return the parameters (e.g., ';foo') for the request.
        """
        return self._params

    def getQueryString( self ):
        """
            Return the query string for the request.
        """
        return self._query

    def getFragment( self ):
        """
            Return the fragment (e.g., '#section') for the request.
        """
        return self._fragment

    def getAuthenticationToken( self ):
        """
            Return the authentication token to be used, if any, as
            a 'userid:password' string.
        """
        return self._authentication_token

    def getHeaders( self ):
        """
            Return a sequence of '( key, value )' tuples for additional
            HTTP headers to be passed with the request.
        """
        return tuple( self._headers )

    def getCookies( self ):
        """
            Return a sequence of '( key, value )' tuples for the cookies
            to be passed with the request.
        """
        return tuple( self._cookies )

    def getFields( self ):
        """
            Return a sequence of '( name, type, value )' tuples
            representing the form variables to be included in the
            request.

            Fields of type 'file' will *not* be loaded;  the value
            will be the filename.
        """
        return tuple( self._fields )

    def getExpectedRedirect( self ):
        """
            Return the URL expected for a redirect.
        """
        return self._expected_redirect

    def getExpectedCookies( self ):
        """
            Return the cookie IDs expected for this request
        """
        return self._expected_cookies

    def matchExpectedCookies( self, cookies ):
        """
            Do 'cookies' match what we expect?
        """
        for cookie in cookies:
            for key in cookie.keys():
                if key not in self._expected_cookies:
                    return 0
        return 1

    def retainCookies( self ):
        """
        """
        return self._expected_cookies is not None
    
    def getContentType( self ):
        """
            Return the HTTP 'Content-type' to use for this request.
        """
        return self._has_file_field and MULTIPART or URLENCODED

    def getData( self ):
        """
            Return the payload for this request, formatted as specified
            in getContentType().
        """
        if self._has_file_field:

            return _mime_encode( self._fields )

        else:

            data = []
            for name, type, value in self._fields:
                value = urllib.quote_plus( value )
                if type:
                    data.append( '%s:%s=%s' % ( name, type, value ) )
                else:
                    data.append( '%s=%s' % ( name, value ) )
            return string.join( data, '&' )

    #
    #   Mutators
    #
    def addHeader( self, header_desc ):
        """
            Construct a header tuple and add it to our list.
        """
        self._headers.append( _buildHeader( header_desc ) )

    def addCookie( self, cookie_desc ):
        """
            Construct a cookie tuple and add it to our list.
        """
        self._cookies.append( _buildCookie( cookie_desc ) )

    def addField( self, field_desc ):
        """
            Construct a field tuple and add it to our list.
        """
        field = _buildField( field_desc )
        if field[2] == 'file':
            self._has_file_field = 1
        self._fields.append( field )

    def setAuthenticationToken( self, authentication_token ):
        """
            Assign the authentication token.
        """
        self._authentication_token = authentication_token

    def setExpectedResult( self, result_code ):
        """
            Override the default expected HTTP response code.
        """
        self._expected_result = result_code

    def setExpectedRedirect( self, redirect_url ):
        """
            Override the default expected HTTP redirect path.
        """
        self._expected_redirect = redirect_url
        if self._expected_result == 200:
            self._expected_result = 302

    def setExpectedTime( self, response_time ):
        """
            Override the default expected response time, in seconds.
        """
        self._expected_time = response_time

    def setExpectedCookies( self, cookies ):
        """
            Override the default expected response time, in seconds.
        """
        if type( cookies ) == type( '' ):
            cookies = cookies.split( ';' )

        self._expected_cookies = cookies

    #
    #   Request invocation
    #
    def __call__( self, result ):
        """
            Invoke the HTTP server with our request;  store the results
            in 'result'.
        """
        invocation = HTTPRequestInvocation( self )

        if result.timeRequests():
            invocation.beginRequest()

        host, port = self.getHost(), self.getPort()
        connection = httplib.HTTP( host, port )
        connection.putrequest( self.getMethod(), self.getURI() )
        connection.putheader( 'Host', '%s:%d' % ( host, port ) )

        token = self.getAuthenticationToken()
        if token:
            scrambled = base64.encodestring( token )
            connection.putheader( 'Authorization', 'Basic %s' % scrambled )

        for key, value in self.getHeaders():
            connection.putheader( key, value )

        for key, value in self.getCookies():
            connection.putheader( 'Cookie', '%s=%s' % ( key, value ) )

        for cookie in result.getCookies():
            connection.putheader( 'Cookie'
                                , cookie.output( header='', attrs=[] ) )

        data = self.getData()
        if data:
            connection.putheader( 'Content-length', str( len( data ) ) )
            connection.putheader( 'Content-type', self.getContentType() )

        connection.endheaders()

        if data:
            connection.send( data )

        code, message, headers = connection.getreply()
        payload = connection.getfile().read()
        invocation.update( code, message, headers, payload )

        if result.timeRequests():
            invocation.endRequest()

        return invocation

def _buildHTTPRequest( cp, section, check_result=1, check_time=1 ):
    """
        Construct an HTTPRequest, using the values in 'cp' and 'section'.
    """
    r = HTTPRequest( section, cp.get( section, 'URL' ) )

    options = cp.options( section )

    if 'authentication' in options:
        r.setAuthenticationToken( cp.get( section, 'authentication' ) )

    header_keys = filter( lambda x, HKEY=HKEY: HKEY.match( x ), options )
    header_keys.sort()

    for hk in header_keys:
        r.addHeader( cp.get( section, hk ) )

    cookie_keys = filter( lambda x, CKEY=CKEY: CKEY.match( x ), options )
    cookie_keys.sort()

    for ck in cookie_keys:
        r.addCookie( cp.get( section, ck ) )

    field_keys = filter( lambda x, FKEY=FKEY: FKEY.match( x ), options )
    field_keys.sort()

    for fk in field_keys:
        r.addField( cp.get( section, fk ) )

    if 'expected_result' in options:
        r.setExpectedResult( cp.getint( section, 'expected_result' ) )

    if 'expected_redirect' in options:
        r.setExpectedRedirect( cp.get( section, 'expected_redirect' ) )

    if 'expected_time' in options:
        r.setExpectedTime( cp.getfloat( section, 'expected_time' ) )

    if 'expected_cookies' in options:
        r.setExpectedCookies( cp.get( section, 'expected_cookies' ) )

    if 'flush_cookies' in options:
        r.setExpectedCookies( None )
    
    return r

class ZEORequest( RequestBase ):
    """
        Represent a scaffolding function call, which will be executed
        via ZEO, rather than via HTTP.
    """

    def __init__( self, name, filename, function ):

        RequestBase.__init__( self, name )

        self._filename = filename
        self._function = function

    #
    #   Accessors
    #
    def getType( self ):
        """
            Return the type, for sniffing purposes.
        """
        return 'ZEORequest'

    def getFileName( self ):
        """
            Return the name of the file containing our function.
        """
        return self._filename

    def getFunctionName( self ):
        """
            Return the name of our function.
        """
        return self._function

    def getFunction( self ):
        """
            Return a function object.
        """
        g = {} # faux globals
        execfile( self.getFileName(), g )
        return g[ self.getFunctionName() ]

    #
    #   Request invocation
    #
    def __call__( self, result ):
        """
            Invoke the function ;  store the results in 'result'.
        """
        invocation  = ZEORequestInvocation( self )
        app         = result.getApplication()
        app._p_jar.sync()   # Ensure that we catch up to any other commits.

        defaults    = result.getDefaults()
        func        = self.getFunction()

        if result.timeRequests():
            invocation.beginRequest()

        invocation.update( func( app, defaults, result ) )

        if result.timeRequests():
            invocation.endRequest()

        return invocation

def _buildZEORequest( cp, section ):
    """
        Construct a ZEORequest, using the values in 'cp' and 'section'.
    """
    r = ZEORequest( section
                  , cp.get( section, 'Filename' )
                  , cp.get( section, 'Function' )
                  )

    options = cp.options( section )

    if 'expected_result' in options:
        r.setExpectedResult( cp.getint( section, 'expected_result' ) )

    if 'expected_time' in options:
        r.setExpectedTime( cp.getfloat( section, 'expected_time' ) )

    return r

def _buildRequest( cp, section ):
    """
        Construct a Request, using the values in 'cp' and 'section'.
    """
    if 'sleep' in cp.options( section ):
        return _buildSleepRequest( cp, section )
    elif 'url' in cp.options( section ):
        return _buildHTTPRequest( cp, section )
    else:
        return _buildZEORequest( cp, section )

#
#   Execution classes
#
class TestBase:
    """
        Interface / common base class for both scenarios ("leaf" tests)
        and suites ("composites").
    """
    _title = ''
    _setup = None
    _teardown = None

    def getTitle( self ):
        """
        """
        return self._title

    def setTitle( self, title ):
        """
        """
        self._title = title

    def setSetup( self, setup ):
        """
        """
        self._setup = setup

    def setTeardown( self, teardown ):
        """
        """
        self._teardown = teardown

    def setup( self, result ):
        if self._setup is not None:
            invocation = self._setup( result )
            if invocation.getResult() != self._setup.getExpectedResult():
                raise RequestError, 'setup failed'

    def teardown( self, result ):
        if self._teardown is not None:
            self._teardown( result )

    #
    #   Plumbing
    #
    def _execute( self, result ):
        """
            Subclass interface:  do whatever, assuming setup has
            succeeded.  Shouldn't raise to indicate "normal" test
            failures.
        """
        raise NotImplemented

    def __call__( self
                , result=None
                , app=None
                , defaults={}
                , time_requests=1
                , check_responses=1
                , check_elapsed_times=1
                ):
        """
            Execute the test, returning result (which we create if
            not passed).
        """
        if result is None:
            result = Result( self   
                           , app
                           , defaults
                           , time_requests
                           , check_responses
                           , check_elapsed_times
                           )
        try:
            self.setup( result )
        except RequestError, err:
            result.logFatalError( err )
            return result

        try:
            try:
                self._execute( result )
            except RequestError, err:
                result.logError( err )
        finally:
            self.teardown( result )

        return result

class Scenario( TestBase ):
    """
    """

    def __init__( self ):
        self._use_case      = ''
        self._requests      = []
        self._postcondition = None
        self._load_sequence = None
        self._load_requests = None

    #
    #   Accessors
    #
    def getUseCase( self ):

        return self._use_case
    
    def getPostcondition( self ):

        return self._postcondition
    
    def listRequests( self ):

        return tuple( self._requests )

    def getLoadSequence( self ):
        
        return self._load_sequence

    def _parseLoadSequence( self ):

        if self._load_requests is None:
            if self._load_sequence:
                self._load_requests = eval( self._load_sequence )

    def hasLoadRequests( self ):

        self._parseLoadSequence()
        return not not self._load_requests

    def nextLoadRequest( self ):

        self._parseLoadSequence()
        lr = self._load_requests

        if lr is None:
            raise IndexError

        result, self._load_requests = lr[0], lr[1:]

        return result

    #
    #   Mutators
    #
    def setUseCase( self, use_case ):

        self._use_case = use_case
    
    def setPostcondition( self, postcondition ):

        self._postcondition = postcondition

    def setLoadSequence( self, load_sequence ):
        """
            Store the string which indicates the sequencing to be
            used as a load-testing script.
        """
        self._load_sequence = load_sequence
    
    def addRequest( self, request, repeat_count=1 ):

        for i in range( repeat_count ):
            self._requests.append( request )

    #
    #   TestBase interface
    #
    def _checkPostcondition( self, result ):
        """
        """
        if self._postcondition is not None:
            pc = self._postcondition
            invocation = pc( result )
            if invocation.getResult() != pc.getExpectedResult():
                raise RequestError, 'failed postcondition'

    def _execute( self, result ):
    
        requests = self.listRequests()

        while result() and requests:

            request, requests = requests[0], requests[1:]
            invocation = request( result )
            if invocation is not None:
                result.addInvocation( invocation, request )

            if hasattr( invocation, 'getCookies' ):
                cookies = invocation.getCookies()
                if cookies and request.matchExpectedCookies( cookies ):
                    result.setCookies( cookies )
                elif not request.retainCookies():
                    result.setCookies( () )

        if result():
            self._checkPostcondition( result )

class Suite( TestBase ):
    """
        Represent a collection of scenarios / sub-suites which make
        up a logical grouping of tests.  Each suite gets a chance to
        do its own setup / teardown, and then delegates running the
        rest of the test to its children.
    """
    def __init__( self ):
        self._children = []

    def listChildren( self ):
        """
        """
        return tuple( self._children )

    def addChild( self, child, repeat_count=1 ):
        """
        """
        for i in range( repeat_count ):
            self._children.append( child )

    def _execute( self, result ):

        for child in self.listChildren():
            childResult = result.newChild( child )
            child( childResult )

SECTION_HEADER = re.compile( r'\[([A-Za-z][A-Za-z0-9_]*)\]')

_SUITE_SECTION          = 'Suite'
_SCENARIO_SECTION       = 'Scenario'
_SETUP_SECTION          = 'Setup'
_POSTCONDITION_SECTION  = 'Postcondition'
_TEARDOWN_SECTION       = 'Teardown'
_DEFAULT_SECTION        = 'DEFAULT'

_SPECIAL_SECTIONS = ( _SUITE_SECTION
                    , _SCENARIO_SECTION
                    , _SETUP_SECTION
                    , _POSTCONDITION_SECTION
                    , _TEARDOWN_SECTION
                    , _DEFAULT_SECTION
                    )

def buildTest( filename, defaults={} ):
    """
        Construct a Suite or Scenario by parsing a file.
    """
    if filename is None:
        raise ValueError, "Must supply 'filename'."

    if '%' in filename:
        filename = filename % defaults

    dir, fn = os.path.split( filename )
    defaults[ 'config_dir' ] = dir
    defaults[ 'config_file' ] = fn

    cp = ConfigParserExt.ConfigParserExt( _SPECIAL_SECTIONS )
    apply( cp.setDefaults, (), defaults )
    cp.read( filename )

    test = None
    isScenario = 0

    if _SUITE_SECTION in cp.listSpecialSections( 0 ):
        isScenario = 0
        test = Suite()
        test.setTitle( cp.get( _SUITE_SECTION, 'Title' ) )

    if _SCENARIO_SECTION in cp.listSpecialSections( 0 ):
        isScenario = 1
        test = Scenario()
        test.setTitle( cp.get( _SCENARIO_SECTION, 'Title' ) )
        test.setUseCase( cp.get( _SCENARIO_SECTION, 'Use_case' ) )

        if cp.has_option( _SCENARIO_SECTION, 'load_sequence' ):
            test.setLoadSequenece( cp.get( _SCENARIO_SECTION
                                 , 'load_sequence' ) )

    if _SETUP_SECTION in cp.listSpecialSections( 0 ):
        test.setSetup( _buildRequest( cp, _SETUP_SECTION ) )

    for section in cp.listOtherSections():
        options = cp.options( section )
        if 'repeat' in options:
            repeat_count = cp.getint( section, 'repeat' )
        else:
            repeat_count = 1
        if isScenario:
            test.addRequest( _buildRequest( cp, section ), repeat_count )
        else:
            child = buildTest( cp.get( section, 'file' ), defaults )
            test.addChild( child, repeat_count )

    if isScenario and _POSTCONDITION_SECTION in cp.listSpecialSections( 0 ):
        test.setPostcondition( _buildRequest( cp, _POSTCONDITION_SECTION ) )

    if _TEARDOWN_SECTION in cp.listSpecialSections( 0 ):
        test.setTeardown( _buildRequest( cp, _TEARDOWN_SECTION ) )

    return test
